@@ -11,12 +11,14 @@
 #include <aconf.h>
 
 #ifdef _WIN32
+#  undef WIN32_LEAN_AND_MEAN
+#  include <windows.h>
 #  include <time.h>
 #  include <direct.h>
+#  include <shobjidl.h>
+#  include <shlguid.h>
 #else
-#  if defined(MACOS)
-#    include <sys/stat.h>
-#  elif !defined(ACORN)
+#  if !defined(ACORN)
 #    include <sys/types.h>
 #    include <sys/stat.h>
 #    include <fcntl.h>
@@ -24,13 +26,15 @@
 #  include <time.h>
 #  include <limits.h>
 #  include <string.h>
-#  if !defined(VMS) && !defined(ACORN) && !defined(MACOS)
+#  if !defined(VMS) && !defined(ACORN)
 #    include <pwd.h>
 #  endif
 #  if defined(VMS) && (__DECCXX_VER < 50200000)
 #    include <unixlib.h>
 #  endif
 #endif // _WIN32
+#include "gmem.h"
+#include "gmempp.h"
 #include "GString.h"
 #include "gfile.h"
 
@@ -47,8 +51,19 @@ GString *getHomeDir() {
   //---------- VMS ----------
   return new GString("SYS$LOGIN:");
 
-#elif defined(__EMX__) || defined(_WIN32)
-  //---------- OS/2+EMX and Win32 ----------
+#elif defined(_WIN32)
+  //---------- Win32 ----------
+  char *s;
+  GString *ret;
+
+  if ((s = getenv("USERPROFILE")))
+    ret = new GString(s);
+  else
+    ret = new GString(".");
+  return ret;
+
+#elif defined(__EMX__)
+  //---------- OS/2+EMX ----------
   char *s;
   GString *ret;
 
@@ -62,10 +77,6 @@ GString *getHomeDir() {
   //---------- RISCOS ----------
   return new GString("@");
 
-#elif defined(MACOS)
-  //---------- MacOS ----------
-  return new GString(":");
-
 #else
   //---------- Unix ----------
   char *s;
@@ -97,8 +108,6 @@ GString *getCurrentDir() {
   if (GetCurrentDirectoryA(sizeof(buf), buf))
 #elif defined(ACORN)
   if (strcpy(buf, "@"))
-#elif defined(MACOS)
-  if (strcpy(buf, ":"))
 #else
   if (getcwd(buf, sizeof(buf)))
 #endif
@@ -179,23 +188,6 @@ GString *appendToPath(GString *path, const char *fileName) {
   }
   return path;
 
-#elif defined(MACOS)
-  //---------- MacOS ----------
-  char *p;
-  int i;
-
-  path->append(":");
-  i = path->getLength();
-  path->append(fileName);
-  for (p = path->getCString() + i; *p; ++p) {
-    if (*p == '/') {
-      *p = ':';
-    } else if (*p == '.') {
-      *p = ':';
-    }
-  }
-  return path;
-
 #elif defined(__EMX__)
   //---------- OS/2+EMX ----------
   int i;
@@ -303,20 +295,12 @@ GString *grabPath(char *fileName) {
     return new GString(fileName, p - fileName);
   return new GString();
 
-#elif defined(MACOS)
-  //---------- MacOS ----------
-  char *p;
-
-  if ((p = strrchr(fileName, ':')))
-    return new GString(fileName, p - fileName);
-  return new GString();
-
 #else
   //---------- Unix ----------
   char *p;
 
   if ((p = strrchr(fileName, '/')))
-    return new GString(fileName, p - fileName);
+    return new GString(fileName, (int)(p - fileName));
   return new GString();
 #endif
 }
@@ -335,10 +319,6 @@ GBool isAbsolutePath(char *path) {
   //---------- RISCOS ----------
   return path[0] == '$';
 
-#elif defined(MACOS)
-  //---------- MacOS ----------
-  return path[0] != ':';
-
 #else
   //---------- Unix ----------
   return path[0] == '/';
@@ -359,11 +339,11 @@ GString *makePathAbsolute(GString *path) {
 
 #elif defined(_WIN32)
   //---------- Win32 ----------
-  char buf[_MAX_PATH];
+  char buf[MAX_PATH];
   char *fp;
 
   buf[0] = '\0';
-  if (!GetFullPathNameA(path->getCString(), _MAX_PATH, buf, &fp)) {
+  if (!GetFullPathNameA(path->getCString(), MAX_PATH, buf, &fp)) {
     path->clear();
     return path;
   }
@@ -376,11 +356,6 @@ GString *makePathAbsolute(GString *path) {
   path->insert(0, '@');
   return path;
 
-#elif defined(MACOS)
-  //---------- MacOS ----------
-  path->del(0, 1);
-  return path;
-
 #else
   //---------- Unix and OS/2+EMX ----------
   struct passwd *pw;
@@ -406,12 +381,12 @@ GString *makePathAbsolute(GString *path) {
 #else
       for (p2 = p1; *p2 && *p2 != '/'; ++p2) ;
 #endif
-      if ((n = p2 - p1) > PATH_MAX)
+      if ((n = (int)(p2 - p1)) > PATH_MAX)
 	n = PATH_MAX;
       strncpy(buf, p1, n);
       buf[n] = '\0';
       if ((pw = getpwnam(buf))) {
-	path->del(0, p2 - p1 + 1);
+	path->del(0, (int)(p2 - p1 + 1));
 	path->insert(0, pw->pw_dir);
       }
     }
@@ -427,6 +402,31 @@ GString *makePathAbsolute(GString *path) {
 #endif
 }
 
+GBool pathIsFile(const char *path) {
+#ifdef _WIN32
+  wchar_t wPath[winMaxLongPath + 1];
+  fileNameToUCS2(path, wPath, winMaxLongPath + 1);
+  DWORD attr = GetFileAttributesW(wPath);
+  return attr != INVALID_FILE_ATTRIBUTES &&
+         !(attr & (FILE_ATTRIBUTE_DIRECTORY | FILE_ATTRIBUTE_DEVICE));
+#else
+  struct stat statBuf;
+  return stat(path, &statBuf) == 0 && S_ISREG(statBuf.st_mode);
+#endif
+}
+
+GBool pathIsDir(const char *path) {
+#ifdef _WIN32
+  wchar_t wPath[winMaxLongPath + 1];
+  fileNameToUCS2(path, wPath, winMaxLongPath + 1);
+  DWORD attr = GetFileAttributesW(wPath);
+  return attr != INVALID_FILE_ATTRIBUTES && (attr & FILE_ATTRIBUTE_DIRECTORY);
+#else
+  struct stat statBuf;
+  return stat(path, &statBuf) == 0 && S_ISDIR(statBuf.st_mode);
+#endif
+}
+
 time_t getModTime(char *fileName) {
 #ifdef _WIN32
   //~ should implement this, but it's (currently) only used in xpdf
@@ -445,25 +445,29 @@ GBool openTempFile(GString **name, FILE **f,
 		   const char *mode, const char *ext) {
 #if defined(_WIN32)
   //---------- Win32 ----------
-  char *tempDir;
+  char tempPath[MAX_PATH + 1];
   GString *s, *s2;
   FILE *f2;
+  DWORD n;
   int t, i;
 
   // this has the standard race condition problem, but I haven't found
   // a better way to generate temp file names with extensions on
   // Windows
-  if ((tempDir = getenv("TEMP"))) {
-    s = new GString(tempDir);
-    s->append('\\');
+  n = GetTempPathA(sizeof(tempPath), tempPath);
+  if (n > 0 && n <= sizeof(tempPath)) {
+    s = new GString(tempPath);
+    if (tempPath[n-1] != '\\') {
+      s->append('\\');
+    }
   } else {
-    s = new GString();
+    s = new GString(".\\");
   }
-  s->appendf("x_{0:d}_{1:d}_",
+  s->appendf("xpdf_{0:d}_{1:d}_",
 	     (int)GetCurrentProcessId(), (int)GetCurrentThreadId());
   t = (int)time(NULL);
   for (i = 0; i < 1000; ++i) {
-    s2 = s->copy()->appendf("{0:d}", t + i);
+    s2 = GString::format("{0:t}{1:d}", s, t + i);
     if (ext) {
       s2->append(ext);
     }
@@ -483,7 +487,7 @@ GBool openTempFile(GString **name, FILE **f,
   }
   delete s;
   return gFalse;
-#elif defined(VMS) || defined(__EMX__) || defined(ACORN) || defined(MACOS)
+#elif defined(VMS) || defined(__EMX__) || defined(ACORN)
   //---------- non-Unix ----------
   char *s;
 
@@ -517,7 +521,7 @@ GBool openTempFile(GString **name, FILE **f,
       *name = new GString("/tmp");
     }
     (*name)->append("/XXXXXX")->append(ext);
-    fd = mkstemps((*name)->getCString(), strlen(ext));
+    fd = mkstemps((*name)->getCString(), (int)strlen(ext));
 #else
     if (!(s = tmpnam(NULL))) {
       return gFalse;
@@ -554,7 +558,7 @@ GBool openTempFile(GString **name, FILE **f,
 
 GBool createDir(char *path, int mode) {
 #ifdef _WIN32
-  return !mkdir(path);
+  return !_mkdir(path);
 #else
   return !mkdir(path, mode);
 #endif
@@ -604,71 +608,120 @@ GString *fileNameToUTF8(wchar_t *path) {
   }
   return s;
 }
+
+GString *fileNameMultiByteToUTF8(char *path) {
+  wchar_t fileNameW[winMaxLongPath + 1];
+  if (MultiByteToWideChar(CP_OEMCP, 0, path, -1,
+			   fileNameW, sizeof(fileNameW) / sizeof(wchar_t))) {
+    return fileNameToUTF8(fileNameW);
+  } else {
+    // shouldn't happen, but just in case...
+    return new GString(path);
+  }
+}
+
+wchar_t *fileNameToUCS2(const char *path, wchar_t *out, size_t outSize) {
+  const char *p;
+  size_t i;
+
+  for (p = path, i = 0; *p && i < outSize - 1; ++i) {
+    if ((p[0] & 0xe0) == 0xc0 &&
+	p[1] && (p[1] & 0xc0) == 0x80) {
+      out[i] = (wchar_t)(((p[0] & 0x1f) << 6) |
+			  (p[1] & 0x3f));
+      p += 2;
+    } else if ((p[0] & 0xf0) == 0xe0 &&
+	       (p[1] & 0xc0) == 0x80 &&
+	       (p[2] & 0xc0) == 0x80) {
+      out[i] = (wchar_t)(((p[0] & 0x0f) << 12) |
+			 ((p[1] & 0x3f) << 6) |
+			  (p[2] & 0x3f));
+      p += 3;
+    } else {
+      out[i] = (wchar_t)(p[0] & 0xff);
+      p += 1;
+    }
+  }
+  out[i] = (wchar_t)0;
+  return out;
+}
 #endif
 
 FILE *openFile(const char *path, const char *mode) {
-#ifdef _WIN32
-  OSVERSIONINFO version;
-  wchar_t wPath[_MAX_PATH + 1];
-  char nPath[_MAX_PATH + 1];
+#if defined(_WIN32)
+  wchar_t wPath[winMaxLongPath + 1];
   wchar_t wMode[8];
-  const char *p;
   int i;
 
-  // NB: _wfopen is only available in NT
-  version.dwOSVersionInfoSize = sizeof(version);
-  GetVersionEx(&version);
-  if (version.dwPlatformId == VER_PLATFORM_WIN32_NT) {
-    for (p = path, i = 0; *p && i < _MAX_PATH; ++i) {
-      if ((p[0] & 0xe0) == 0xc0 &&
-	  p[1] && (p[1] & 0xc0) == 0x80) {
-	wPath[i] = (wchar_t)(((p[0] & 0x1f) << 6) |
-			     (p[1] & 0x3f));
-	p += 2;
-      } else if ((p[0] & 0xf0) == 0xe0 &&
-		 p[1] && (p[1] & 0xc0) == 0x80 &&
-		 p[2] && (p[2] & 0xc0) == 0x80) {
-	wPath[i] = (wchar_t)(((p[0] & 0x0f) << 12) |
-			     ((p[1] & 0x3f) << 6) |
-			     (p[2] & 0x3f));
-	p += 3;
-      } else {
-	wPath[i] = (wchar_t)(p[0] & 0xff);
-	p += 1;
-      }
-    }
-    wPath[i] = (wchar_t)0;
-    for (i = 0; mode[i] && i < sizeof(mode) - 1; ++i) {
-      wMode[i] = (wchar_t)(mode[i] & 0xff);
-    }
-    wMode[i] = (wchar_t)0;
-    return _wfopen(wPath, wMode);
-  } else {
-    for (p = path, i = 0; *p && i < _MAX_PATH; ++i) {
-      if ((p[0] & 0xe0) == 0xc0 &&
-	  p[1] && (p[1] & 0xc0) == 0x80) {
-	nPath[i] = (char)(((p[0] & 0x1f) << 6) |
-			  (p[1] & 0x3f));
-	p += 2;
-      } else if ((p[0] & 0xf0) == 0xe0 &&
-		 p[1] && (p[1] & 0xc0) == 0x80 &&
-		 p[2] && (p[2] & 0xc0) == 0x80) {
-	nPath[i] = (char)(((p[1] & 0x3f) << 6) |
-			  (p[2] & 0x3f));
-	p += 3;
-      } else {
-	nPath[i] = p[0];
-	p += 1;
-      }
-    }
-    nPath[i] = '\0';
-    return fopen(nPath, mode);
+  fileNameToUCS2(path, wPath, winMaxLongPath + 1);
+  for (i = 0; mode[i] && i < sizeof(wMode)/sizeof(wchar_t) - 1; ++i) {
+    wMode[i] = (wchar_t)(mode[i] & 0xff);
   }
+  wMode[i] = (wchar_t)0;
+  readWindowsShortcut(wPath, winMaxLongPath + 1);
+  return _wfopen(wPath, wMode);
+#elif defined(VMS)
+  return fopen(path, mode, "ctx=stm");
 #else
   return fopen(path, mode);
 #endif
 }
 
+#ifdef _WIN32
+void readWindowsShortcut(wchar_t *wPath, size_t wPathSize) {
+  size_t n = wcslen(wPath);
+  if (n < 4 || wcscmp(wPath + n - 4, L".lnk")) {
+    return;
+  }
+  IShellLinkW *shellLink;
+  HRESULT hres;
+  hres = CoCreateInstance(CLSID_ShellLink, NULL, CLSCTX_INPROC_SERVER,
+			  IID_IShellLinkW, (LPVOID *)&shellLink);
+  bool needCoUninit = false;
+  if (hres == CO_E_NOTINITIALIZED) {
+    CoInitialize(NULL);
+    needCoUninit = true;
+    hres = CoCreateInstance(CLSID_ShellLink, NULL, CLSCTX_INPROC_SERVER,
+			    IID_IShellLinkW, (LPVOID *)&shellLink);
+  }
+  if (FAILED(hres)) {
+    return;
+  }
+  IPersistFile *persistFile;
+  hres = shellLink->QueryInterface(IID_IPersistFile, (LPVOID *)&persistFile);
+  if (FAILED(hres)) {
+    return;
+  }
+  hres = persistFile->Load(wPath, STGM_READ);
+  if (FAILED(hres)) {
+    fprintf(stderr, "IPersistFile.Load failed: 0x%08lx\n", hres);
+    exit(1);
+  }
+  wchar_t target[winMaxLongPath + 1];
+  hres = shellLink->GetPath(target, winMaxLongPath + 1, NULL, 0);
+  if (FAILED(hres)) {
+    return;
+  }
+  shellLink->Release();
+  if (needCoUninit) {
+    CoUninitialize();
+  }
+  if (wcslen(target) > wPathSize - 1) {
+    return;
+  }
+  wcscpy(wPath, target);
+}
+#endif
+
+int makeDir(const char *path, int mode) {
+#ifdef _WIN32
+  wchar_t wPath[winMaxLongPath + 1];
+  return _wmkdir(fileNameToUCS2(path, wPath, winMaxLongPath + 1));
+#else
+  return mkdir(path, (mode_t)mode);
+#endif
+}
+
 char *getLine(char *buf, int size, FILE *f) {
   int c, i;
 
@@ -722,151 +775,28 @@ GFileOffset gftell(FILE *f) {
 #endif
 }
 
-//------------------------------------------------------------------------
-// GDir and GDirEntry
-//------------------------------------------------------------------------
-
-GDirEntry::GDirEntry(char *dirPath, char *nameA, GBool doStat) {
-#ifdef VMS
-  char *p;
-#elif defined(_WIN32)
-  int fa;
-  GString *s;
-#elif defined(ACORN)
-#else
-  struct stat st;
-  GString *s;
-#endif
-
-  name = new GString(nameA);
-  dir = gFalse;
-  if (doStat) {
-#ifdef VMS
-    if (!strcmp(nameA, "-") ||
-	((p = strrchr(nameA, '.')) && !strncmp(p, ".DIR;", 5)))
-      dir = gTrue;
-#elif defined(ACORN)
-#else
-    s = new GString(dirPath);
-    appendToPath(s, nameA);
+void fixCommandLine(int *argc, char **argv[]) {
 #ifdef _WIN32
-    fa = GetFileAttributesA(s->getCString());
-    dir = (fa != 0xFFFFFFFF && (fa & FILE_ATTRIBUTE_DIRECTORY));
-#else
-    if (stat(s->getCString(), &st) == 0)
-      dir = S_ISDIR(st.st_mode);
-#endif
-    delete s;
-#endif
+  int argcw;
+  wchar_t **argvw;
+  GString *arg;
+  int i;
+
+  argvw = CommandLineToArgvW(GetCommandLineW(), &argcw);
+  if (!argvw || argcw < 0) {
+    return;
   }
-}
 
-GDirEntry::~GDirEntry() {
-  delete name;
-}
+  *argc = argcw;
 
-GDir::GDir(char *name, GBool doStatA) {
-  path = new GString(name);
-  doStat = doStatA;
-#if defined(_WIN32)
-  GString *tmp;
-
-  tmp = path->copy();
-  tmp->append("/*.*");
-  hnd = FindFirstFileA(tmp->getCString(), &ffd);
-  delete tmp;
-#elif defined(ACORN)
-#elif defined(MACOS)
-#elif defined(ANDROID)
-#else
-  dir = opendir(name);
-#ifdef VMS
-  needParent = strchr(name, '[') != NULL;
-#endif
-#endif
-}
-
-GDir::~GDir() {
-  delete path;
-#if defined(_WIN32)
-  if (hnd) {
-    FindClose(hnd);
-    hnd = NULL;
+  *argv = (char **)gmallocn(argcw + 1, sizeof(char *));
+  for (i = 0; i < argcw; ++i) {
+    arg = fileNameToUTF8(argvw[i]);
+    (*argv)[i] = copyString(arg->getCString());
+    delete arg;
   }
-#elif defined(ACORN)
-#elif defined(MACOS)
-#elif defined(ANDROID)
-#else
-  if (dir)
-    closedir(dir);
-#endif
-}
+  (*argv)[argcw] = NULL;
 
-GDirEntry *GDir::getNextEntry() {
-  GDirEntry *e;
-
-#if defined(_WIN32)
-  if (hnd) {
-    e = new GDirEntry(path->getCString(), ffd.cFileName, doStat);
-    if (hnd  && !FindNextFileA(hnd, &ffd)) {
-      FindClose(hnd);
-      hnd = NULL;
-    }
-  } else {
-    e = NULL;
-  }
-#elif defined(ACORN)
-#elif defined(MACOS)
-#elif defined(ANDROID)
-#elif defined(VMS)
-  struct dirent *ent;
-  e = NULL;
-  if (dir) {
-    if (needParent) {
-      e = new GDirEntry(path->getCString(), "-", doStat);
-      needParent = gFalse;
-      return e;
-    }
-    ent = readdir(dir);
-    if (ent) {
-      e = new GDirEntry(path->getCString(), ent->d_name, doStat);
-    }
-  }
-#else
-  struct dirent *ent;
-  e = NULL;
-  if (dir) {
-    ent = (struct dirent *)readdir(dir);
-    if (ent && !strcmp(ent->d_name, ".")) {
-      ent = (struct dirent *)readdir(dir);
-    }
-    if (ent) {
-      e = new GDirEntry(path->getCString(), ent->d_name, doStat);
-    }
-  }
-#endif
-
-  return e;
-}
-
-void GDir::rewind() {
-#ifdef _WIN32
-  GString *tmp;
-
-  if (hnd)
-    FindClose(hnd);
-  tmp = path->copy();
-  tmp->append("/*.*");
-  hnd = FindFirstFileA(tmp->getCString(), &ffd);
-  delete tmp;
-#elif defined(ACORN)
-#elif defined(MACOS)
-#elif defined(ANDROID)
-#else
-  if (dir)
-    rewinddir(dir);
-#ifdef VMS
-  needParent = strchr(path->getCString(), '[') != NULL;
-#endif
+  LocalFree(argvw);
 #endif
 }
